# -*- coding: utf-8 -*-
import os, sys, errno
from stat import S_ISCHR, S_ISFIFO, S_ISSOCK
from subprocess import TimeoutExpired, PIPE, Popen
from socket import socketpair as create_socket_pair

from .futures import Task, Future, WaitTillAll, future_or_timeout
from .protocol import ReadProtocolBase

IS_AIX = sys.platform.startswith('aix')
LIMIT = 1<<16
MAX_READ_SIZE = 262144

PROCESS_EXIT_DELAY_LIMIT = 10

class UnixReadPipeTransport:
    """
    Asynchronous read only transport implementation for pipes.
    
    Attributes
    ----------
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    closing : `bool`
        Whether the transport ic closing.
    fileno : `int`
        The used socket's file descriptor number.
    loop : ``EventThread``
        The respective event loop of the transport.
    pipe : `None` or `file-like` object
        The pipe to connect to on read end.
        
        Is set to non-blocking mode.
        
        After closing the transport is set to `None`.
    protocol : `None`, ``SubprocessReadPipeProtocol``, `Any`
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    """
    __slots__ = ('_extra', '_paused', 'closing', 'fileno', 'loop', 'pipe', 'protocol')
    async def __new__(cls, loop, pipe, protocol, extra=None):
        """
        Creates a new ``UnixReadPipeTransport`` instance with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop of the transport.
        pipe : `file-like` object
            The pipe to connect to on read end.
        protocol : ``SubprocessReadPipeProtocol`` or `Any`
            Asynchronous protocol implementation used by the transport.
        extra : `None` or `dict` of (`str`, `Any`) items, Optional
            Optional transport information.
        
        Raises
        ------
        ValueError
            If `pipe` was not given neither as pipe, socket or character device.
        """
        fileno = pipe.fileno()
        mode = os.fstat(fileno).st_mode
        if not (S_ISFIFO(mode) or S_ISSOCK(mode) or S_ISCHR(mode)):
            raise ValueError(f'{cls.__name__} is only for pipes, sockets and character devices, got '
                f'{pipe.__class__.__name__}; {pipe!r}.')
        
        self = object.__new__(cls)
        if extra is None:
            extra = {}
        
        extra['pipe'] = pipe
        
        self._extra = extra
        self.loop = loop
        self.pipe = pipe
        self.fileno = fileno
        self.protocol = protocol
        self.closing = False
        self._paused = False
        
        try:
            os.set_blocking(fileno, False)
            # skip 1 callback loop
            future = Future(loop)
            loop.call_soon(Future.set_result_if_pending, future, None)
            await future
            
            protocol.connection_made(self)
            loop.add_reader(fileno, self._read_ready)
        
        except:
            self.close()
            raise
        
        return self
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self._extra.get(name, default)
    
    def __repr__(self):
        """Returns the ``UnixReadPipeTransport``'s representation."""
        return f'<{self.__class__.__name__} fd={self.fileno}>'
    
    def _read_ready(self):
        """
        Added as a read callback on the respective event loop to be called when the data is received on the pipe.
        """
        try:
            data = os.read(self.fileno, MAX_READ_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as err:
            self._fatal_error(err, 'Fatal read error on pipe transport')
        else:
            if data:
                self.protocol.data_received(data)
            else:
                self.closing = True
                loop = self.loop
                loop.remove_reader(self.fileno)
                protocol = self.protocol
                loop.call_soon(protocol.__class__.eof_received, protocol)
                loop.call_soon(self.__class__._call_connection_lost, self, None)
    
    def pause_reading(self):
        """
        Pauses the receiving end.
        
        No data will be passed to the respective protocol's ``.data_received`` method until ``.resume_reading`` is
        called.
        
        Returns
        -------
        reading_paused : `bool`
            Whether reading was paused.
        """
        if self.closing or self._paused:
            return False
        
        self._paused = True
        self.loop.remove_reader(self.fileno)
        return True
    
    def resume_reading(self):
        """
        Resumes the receiving end.
        
        Data received will once again be passed to the respective protocol's ``.data_received`` method.
        
        Returns
        -------
        reading_resume : `bool`
            Whether reading was resumed.
        """
        if self.closing or not self._paused:
            return False
        
        self._paused = False
        self.loop.add_reader(self.fileno, self._read_ready)
        return True
    
    def set_protocol(self, protocol):
        """
        Sets a new protocol to the transport.
        
        Parameters
        ----------
        protocol : ``SubprocessReadPipeProtocol`` or `Any`
            Asynchronous protocol implementation.
        """
        self.protocol = protocol
    
    def get_protocol(self):
        """
        Gets the read pipe transport's actual protocol.
        
        Returns
        -------
        protocol : `None`, ``SubprocessReadPipeProtocol`` or `Any`
            Asynchronous protocol implementation.
        """
        return self.protocol
    
    def is_closing(self):
        """
        Returns whether the read pipe transport is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self.closing
    
    def close(self):
        """
        Starts the shutdown process of the read pipe transport.
        """
        if not self.closing:
            self._close(None)
    
    def __del__(self):
        """
        Closes the read pipe transport if not yet closed.
        """
        pipe = self.pipe
        if (pipe is not None):
            pipe.close()
    
    def _fatal_error(self, exception, message='Fatal error on pipe transport'):
        """
        If a fatal error occurs on the transport, renders its traceback and closes itself.
        
        Parameters
        ----------
        exception : `BaseException`
            The occurred exception.
        message : `str`, Optional
            Additional error message to render.
        """
        if not (isinstance(exception, OSError) and (exception.errno == errno.EIO)):
            self.loop.render_exc_async(exception, [message, '\non: ', repr(self), '\n'])
        
        self._close(exception)
    
    def _close(self, exception):
        """
        Starts the transport's closing process.
        
        Parameters
        ----------
        exception : `None` or ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        self.closing = True
        loop = self.loop
        loop.remove_reader(self.fileno)
        loop.call_soon(self.__class__._call_connection_lost, self, exception)
    
    def _call_connection_lost(self, exception):
        """
        Calls the read pipe transport's protocol's `.connection_lost` with the given exception and closes the
        transport's pipe.
        
        Parameters
        ----------
        exception : `None` or ``BaseException``
            Exception to call the protocol's ``.connection_lost`` with.
            
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        protocol = self.protocol
        if protocol is None:
            return
        
        try:
            protocol.connection_lost(exception)
        finally:
            pipe = self.pipe
            if (pipe is not None):
                self.pipe = None
                pipe.close()
            
            self.protocol = None

class UnixWritePipeTransport:
    """
    Asynchronous write only transport implementation for pipes.
    
    Attributes
    ----------
    _buffer : `bytearray`
        Data ensured to be written on the wrapped pipe as it becomes readable again.
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _high_water : `int`
        The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
    _low_water : `int`
        The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to `16384`.
    closing : `bool`
        Whether the transport ic closing.
    fileno : `int`
        The used socket's file descriptor number.
    loop : ``EventThread``
        The respective event loop of the transport.
    pipe : `None` or `file-like` object
        The pipe to connect to on read end.
        
        Is set to non-blocking mode.
        
        After closing the transport is set to `None`.
    protocol : `None`, ``SubprocessWritePipeProtocol`` or `Any`
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    protocol_paused : `bool`
        Whether ``.protocol`` is paused writing.
    """
    __slots__ = ('_buffer', '_extra', '_high_water', '_low_water', 'closing', 'fileno', 'loop', 'pipe', 'protocol',
        'protocol_paused')
    
    async def __new__(cls, loop, pipe, protocol, extra=None):
        """
        Creates a new ``UnixWritePipeTransport`` instance with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop of the transport.
        pipe `: file-like` object
            The pipe to connect to on read end.
        protocol : ``SubprocessWritePipeProtocol`` or `Any`
            Asynchronous protocol implementation used by the transport.
        extra : `None` or `dict` of (`str`, `Any`) items, Optional
            Optional transport information.

        Raises
        ------
        ValueError
            If `pipe` was not given neither as pipe, socket or character device.
        """
        fileno = pipe.fileno()
        mode = os.fstat(fileno).st_mode
        is_char = S_ISCHR(mode)
        is_fifo = S_ISFIFO(mode)
        is_socket = S_ISSOCK(mode)
        if not (is_char or is_fifo or is_socket):
            raise ValueError('Pipe transport is only for pipes, sockets and character devices, got '
                f'{pipe.__class__.__name__}; {pipe!r}.')
        
        if extra is None:
            extra = {}
        
        extra['pipe'] = pipe
        
        self = object.__new__(cls)
        self._extra = extra
        self.loop = loop
        self.protocol_paused = False
        self.pipe = pipe
        self.fileno = fileno
        self.protocol = protocol
        self._buffer = bytearray()
        self.closing = False  # Set when close() or write_eof() called.
        
        self._high_water = 65536
        self._low_water = 16384
        
        try:
            os.set_blocking(fileno, False)
            # skip 1 callback loop
            future = Future(loop)
            loop.call_soon(Future.set_result_if_pending, future, None)
            await future
            
            protocol.connection_made(self)
            
            # On AIX, the reader trick (to be notified when the read end of the  socket is closed) only works for
            # sockets. On other platforms it works for pipes and sockets.
            if is_socket or (is_fifo and not IS_AIX):
                loop.add_reader(fileno, self._read_ready)
        except:
            self.close()
            raise
        
        return self
    
    def __repr__(self):
        """Returns the ``UnixWritePipeTransport``'s representation."""
        return f'<{self.__class__.__name__} fd={self.fileno}>'
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self._extra.get(name, default)
    
    def get_write_buffer_size(self):
        """
        Return the current size of the write buffer.
        
        Returns
        -------
        get_write_buffer_size : `int`
        """
        return len(self._buffer)
    
    def _read_ready(self):
        """
        Added as a read callback on the respective event loop to be called when the data is received on the pipe.
        
        If this happens, since it is a write only pipe, means it should be closed, so we do like that.
        """
        # Pipe was closed by peer.
        if self._buffer:
            exception = BrokenPipeError()
        else:
            exception = None
            
        self._close(exception)
    
    def write(self, data):
        """
        Write the given data to the transport.
        
        The method do no blocks, instead arranges the data to be sent asynchronously.
        
        Parameters
        ----------
        data : `bytes-like`
            The bytes data to be sent.
        """
        if not data:
            return
        
        if isinstance(data, bytearray):
            data = memoryview(data)
        
        if self.closing:
            return
        
        buffer = self._buffer
        if not buffer:
            try:
                n = os.write(self.fileno, data)
            except (BlockingIOError, InterruptedError):
                n = 0
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on pipe transport')
                return
            
            if n == len(data):
                return
            
            if n > 0:
                data = memoryview(data)[n:]
            
            self.loop.add_writer(self.fileno, self._write_ready)
        
        buffer.extend(data)
        self._maybe_pause_protocol()
    
    def _write_ready(self):
        """
        Added as a write callback on the respective event loop when the transport has unsent data. Called when the
        respective socket becomes writable.
        """
        buffer = self._buffer
        
        try:
            n = os.write(self.fileno, buffer)
        except (BlockingIOError, InterruptedError):
            pass
        except BaseException as err:
            buffer.clear()
            self.loop.remove_writer(self.fileno)
            self._fatal_error(err, 'Fatal write error on pipe transport')
        else:
            if n == len(buffer):
                buffer.clear()
                self.loop.remove_writer(self.fileno)
                self._maybe_resume_protocol()  # May append to buffer.
                if self.closing:
                    self.loop.remove_reader(self.fileno)
                    self._call_connection_lost(None)
                return
            
            if n > 0:
                del buffer[:n]
    
    def can_write_eof(self):
        """
        Return whether the transport supports ``.write_eof``.
        
        Returns
        -------
        can_write_eof : `bool`
            ``UnixWritePipeTransport`` instances always return `True`.
        """
        return True
    
    def write_eof(self):
        """
        Writes eof to the transport's protocol if applicable.
        
        If the write transport's buffer is empty, calls connection lost as well.
        """
        if self.closing:
            return
        
        self.closing = True
        if not self._buffer:
            loop = self.loop
            loop.remove_reader(self.fileno)
            loop.call_soon(self.__class__._call_connection_lost, self, None)
    
    def set_protocol(self, protocol):
        """
        Sets a new protocol to the transport.
        
        Parameters
        ----------
        protocol : ``SubprocessWritePipeProtocol`` or `Any`
            Asynchronous protocol implementation.
        """
        self.protocol = protocol
    
    def get_protocol(self):
        """
        Gets the transport's actual protocol.
        
        Returns
        -------
        protocol : `None`, ``SubprocessWritePipeProtocol` or  `Any`
            Asynchronous protocol implementation.
        """
        return self.protocol
    
    def is_closing(self):
        """
        Returns whether the read pipe transport is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self.closing
    
    def close(self):
        """
        Starts the shutdown process of the write pipe transport.
        """
        if (self.pipe is not None) and (not self.closing):
            self.write_eof()
    
    def __del__(self):
        """
        Closes the write pipe transport if not yet closed.
        """
        pipe = self.pipe
        if (pipe is not None):
            pipe.close()
    
    def abort(self):
        """
        Close the transport immediately.
        
        The buffered data will be lost.
        """
        self._close(None)
    
    def _fatal_error(self, exception, message='Fatal error on pipe transport'):
        """
        If a fatal error occurs on the transport, renders its traceback and closes itself.
        
        Parameters
        ----------
        exception : `BaseException`
            The occurred exception.
        message : `str`, Optional
            Additional error message to render.
        """
        if not isinstance(exception, OSError):
            self.loop.render_exc_async(exception, [message, '\non: ', repr(self), '\n'])
        
        self._close(exception)
    
    def _close(self, exception):
        """
        Starts the transport's closing process.
        
        Parameters
        ----------
        exception : `None` or ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        self.closing = True
        
        loop = self.loop
        buffer = self._buffer
        if buffer:
            self.loop.remove_writer(self.fileno)
            buffer.clear()
        
        loop.remove_reader(self.fileno)
        loop.call_soon(self.__class__._call_connection_lost, self, exception)
    
    def _call_connection_lost(self, exception):
        """
        Calls the write pipe transport's protocol's `.connection_lost` with the given exception and closes the
        transport's pipe.
        
        Parameters
        ----------
        exception : `None` or ``BaseException``
            Exception to call the protocol's ``.connection_lost`` with.
            
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        protocol = self.protocol
        if protocol is None:
            return
        
        try:
            protocol.connection_lost(exception)
        finally:
            pipe = self.pipe
            if (pipe is not None):
                self.pipe = None
                pipe.close()
            
            self.protocol = None
    
    def _maybe_pause_protocol(self):
        """
        Called after data was ensured to be written into the pipe to check whether it's protocol should be paused.
        """
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        
        if self.protocol_paused:
            return
        
        self.protocol_paused = True
        
        protocol = self.protocol
        if protocol is None:
            return
        
        try:
            protocol.pause_writing()
        except BaseException as err:
            self.loop.render_exc_async(err, [
                repr(self), '`._maybe_pause_protocol` failed\n'
                'On: ', repr(protocol), '.pause_writing()\n'])
    
    def _maybe_resume_protocol(self):
        """
        Called after successful writing to the pipe to check whether the protocol should be resumed.
        """
        if (self.protocol_paused and self.get_write_buffer_size() <= self._low_water):
            self.protocol_paused = False
            protocol = self.protocol
            if (protocol is not None):
                try:
                    protocol.resume_writing()
                except BaseException as err:
                    self.loop.render_exc_async(err, [
                        repr(self), '`._maybe_resume_protocol` failed\n'
                        'on: ', repr(protocol), '.resume_writing()\n'])
    
    def get_write_buffer_limits(self):
        """
        Returns the low and the high water of the transport.
        
        Returns
        -------
        low_water : `int`
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
        high_water : `int`
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.
        """
        return (self._low_water, self._high_water)
    
    def _set_write_buffer_limits(self, low=None, high=None):
        """
        Sets the write buffer limits of the transport.
        
        Parameters
        ----------
        low : None` or `int`, Optional
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
        high : `None` or `int`, Optional
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.

        Raises
        ------
        ValueError
            If `high` is lower than `low` or if `low` is lower than `0`.
        """
        if high is None:
            if low is None:
                high = 65536
                low = 16384
            else:
                high = low<<2
        else:
            if low is None:
                low = high>>2
        
        if low < 0 or high < low:
            raise ValueError(f'High water must be greater or equal than low, what must be greater than equal than `0`, '
                f'got high={high!r}; low={low!r}.')
        
        self._high_water = high
        self._low_water = low
    
    def set_write_buffer_limits(self, low=None, high=None):
        """
        Set the high- and low-water limits for write flow control.
        
        These two values control when to call the protocol's ``.pause_writing`` and ``.resume_writing`` methods. If
        specified, the low-water limit must be less than or equal to the high-water limit. Neither value can be
        negative. The defaults are implementation-specific. If only the high-water limit is given, the low-water limit
        defaults to an implementation-specific value less than or equal to the high-water limit. Setting high to zero
        forces low to zero as well, and causes ``.pause_writing`` to be called whenever the buffer becomes non-empty.
        Setting low to zero causes ``.resume_writing`` to be called only once the buffer is empty. Use of zero for
        either limit is generally sub-optimal as it reduces opportunities for doing I/O and computation concurrently.
        
        Parameters
        ----------
        low : None` or `int`, Optional
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
        high : `None` or `int`, Optional
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.
        
        Raises
        ------
        ValueError
            If `low` is lower than `0` or if `low` is higher than `high`.
        """
        self._set_write_buffer_limits(low=low, high=high)
        self._maybe_pause_protocol()

class SubprocessStreamWriter:
    """
    Writer interface for subprocess calls.
    
    Attributes
    ----------
    loop : ``EventThread``
        The respective event loop of the stream.
    transport : ``UnixWritePipeTransport`` or `Any`
        Asynchronous transport implementation.
    protocol : ``AsyncProcess`` or `Any`
        Asynchronous protocol implementation.
    """
    __slots__ = ('loop', 'transport', 'protocol', )
    def __init__(self, loop, transport, protocol):
        """
        Stream-writer used as ``AsyncProcess.stdin``.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop of the stream.
        transport : ``UnixWritePipeTransport`` or `Any`
            Asynchronous transport implementation.
        protocol : ``AsyncProcess`` or `Any`
            Asynchronous protocol implementation.
        """
        self.transport = transport
        self.protocol = protocol
        self.loop = loop
    
    def write(self, data):
        """
        Write the given data to subprocess pipe's transport.
        
        The method do no blocks, instead arranges the data to be sent asynchronously.
        
        Parameters
        ----------
        data : `bytes-like`
            The bytes data to be sent.
        """
        self.transport.write(data)
    
    def writelines(self, lines):
        """
        Writes the given lines to the subprocess pipe's transport.
        
        Parameters
        ----------
        lines : `iterable` of `bytes-like`
            The lines to write.
        
        Raises
        ------
        RuntimeError
            Protocol has no attached transport.
        """
        self.transport.writelines(lines)
    
    def write_eof(self):
        """
        Writes eof to the subprocess pipe's transport's protocol if applicable.
        
        By default ``SubprocessStreamWriter``'s transport is ``UnixWritePipeTransport``, what will call connection lost
        as well when the write buffer is empty.
        """
        return self.transport.write_eof()
    
    def can_write_eof(self):
        """
        Return whether the pipe's transport supports ``.write_eof``.
        
        Returns
        -------
        can_write_eof : `bool`
            By default ``SubprocessStreamWriter``'s transport is ``UnixWritePipeTransport`` instance, what always
            return `True`.
        """
        return self.transport.can_write_eof()
    
    def close(self):
        """
        Starts the shutdown process of subprocess pipe's transport.
        """
        return self.transport.close()
    
    def is_closing(self):
        """
        Returns whether the subprocess pipe's transport is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self.transport.is_closing()
    
    async def wait_closed(self):
        """
        Blocks till the subprocess pipe's protocol closes.
        
        This method is a coroutine.
        """
        await self.protocol._get_close_waiter(self)
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self.transport.get_extra_info(name, default)
    
    async def drain(self):
        """
        Blocks till the write buffer is drained.
        
        This method is a coroutine.
        
        Raises
        ------
        BaseException
            Connection lost exception if applicable.
        """
        if self.transport.is_closing():
            loop = self.loop
            future = Future(loop)
            loop.call_soon(Future.set_result_if_pending, future, None)
            await future
        
        await self.protocol._drain_helper()

class SubprocessWritePipeProtocol:
    """
    Asynchronous subprocess write pipe protocol.
    
    Attributes
    ----------
    disconnected : `bool`
        Whether the protocol is disconnected.
    fd : `int`
        The used socket's file descriptor number.
    transport : ``UnixWritePipeTransport`` or `Any`
        Asynchronous transport implementation.
    process : ``AsyncProcess``
        The parent process of the pipe protocol.
    """
    __slots__ = ('disconnected', 'fd', 'transport', 'process', )
    
    def __init__(self, process, fd):
        """
        Creates a new ``SubprocessWritePipeProtocol`` instance with the given parameters.
        
        Parameters
        ----------
        process : ``AsyncProcess``
            The parent process of the pipe protocol.
        fd : `int`
            The used socket's file descriptor number.
        """
        self.process = process
        self.fd = fd
        self.transport = None
        self.disconnected = False
    
    def __call__(self):
        """
        ``SubprocessWritePipeProtocol`` instances return themselves allowing using them as their one-time factory.
        
        Returns
        -------
        self : ``SubprocessWritePipeProtocol``
        """
        return self
    
    def connection_made(self, transport):
        """
        Called when the connection is made.
        
        Sets the ``SubprocessWritePipeProtocol``'s ``.transport``.
        
        Parameters
        ----------
        transport : ``UnixWritePipeTransport`` or `Any`
            Asynchronous transport implementation.
        """
        self.transport = transport
    
    def __repr__(self):
        """Returns the subprocess write protocol's representation."""
        return f'<{self.__class__.__name__} fd={self.fd} pipe={self.transport!r}>'
    
    def connection_lost(self, exception):
        """
        Called when the connection is lost or closed.
        
        Parameters
        ----------
        exception : `None` or `BaseException` instance
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        process = self.process
        if (process is not None):
            self.process = None
            process._pipe_connection_lost(self.fd, exception)
            self.disconnected = True
    
    def pause_writing(self):
        """
        Called when the transport's buffer goes over the high-water mark.
        
        ``.pause_writing`` is called when the buffer goes over the high-water mark, and eventually
        ``.resume_writing`` is called when the buffer size reaches the low-water mark.
        """
        self.process.pause_writing()
    
    def resume_writing(self):
        """
        Called when the transport's buffer drains below the low-water mark.
        
        See ``.pause_writing`` for details.
        """
        self.process.resume_writing()

class SubprocessReadPipeProtocol(SubprocessWritePipeProtocol):
    """
    Asynchronous subprocess read pipe protocol.
    
    Attributes
    ----------
    disconnected : `bool`
        Whether the protocol is disconnected.
    fd : `int`
        The used socket's file descriptor number.
    transport : ``UnixWritePipeTransport`` or `Any`
        Asynchronous transport implementation.
    process : ``AsyncProcess``
        The parent process of the pipe protocol.
    """
    __slots__ = ()
    def data_received(self, data):
        """
        Called when some data is received on the pipe.
        
        Parameters
        ----------
        data : `bytes`
            The received data.
        """
        self.process._pipe_data_received(self.fd, data)
    
    def eof_received(self):
        # connection_lost is received anyways after eof, and connection_lost without exception will cause eof.
        """
        Calling ``.connection_lost`` without exception causes eof. ``.connection_lost`` is called anyways after
        ``.eof_received``, so this method does nothing.
        
        Returns
        -------
        transport_closes : `bool`
            Returns `False` if the transport will close itself. If it returns `True`, then closing the transport is up
            to the protocol.
            
            Always returns `False`.
        """
        return False


class AsyncProcess:
    """
    Asynchronous process implementation.
    
    Attributes
    ----------
    _alive_fds : `list` if `int`
        A list of alive file descriptor's respective internal identifier.
        
        Can have the following elements:
        
        +-------------------+-------+
        | Respective name   | Value |
        +===================+=======+
        | stdin             | `0`   |
        +-------------------+-------+
        | stdout            | `1`   |
        +-------------------+-------+
        | stderr            | `2`   |
        +-------------------+-------+
    
    _connection_lost : `bool`
        Whether all the pipes of the ``AsyncProcess`` lost connection.
    _drain_waiter : `None` or ``Future``
        A future, what is used to block the writing task, till it's writen data is drained.
    _exit_waiters : `None` or `set of ``Future``
        Waiter futures which wait for the subprocess to shutdown.
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _paused : `bool`
        Whether the subprocess is paused writing because it hit's the high water mark.
    _pending_calls : `None` or `list` of (`callable`, `tuple` of `Any`)
        Meanwhile the subprocess connection is established, this attribute is set as a list to put connection lost
        and related calls it to with their parameters.
    _subprocess_stderr_protocol : `None` or ``SubprocessReadPipeProtocol``
        Protocol of the stderr pipe if applicable.
    _subprocess_stdin_protocol : `None` or ``SubprocessWritePipeProtocol``
        Protocol of the stdin pipe if applicable.
    _subprocess_stdout_protocol : `None` or ``SubprocessReadPipeProtocol``
        Protocol of the stdout pipe if applicable.
    closed : `bool`
        Whether subprocess is closed.
    loop : ``EventThread``
        the respective event loop of the async subprocess to what is bound to.
    process : `subprocess.Process`
        The internal blocking subprocess object.
    process_id : `int`
        The subprocess identifier.
    return_code : `None` or `int`
        The returned exit code of the subprocess. Set as `None` if not yet applicable.
    stderr : ``ReadProtocolBase``
        Asynchronous stderr implementation.
    stdin : ``SubprocessStreamWriter``
        Asynchronous stdin implementation.
    stdout : ``ReadProtocolBase``
        Asynchronous stdout implementation.
    """
    __slots__ = ('_alive_fds', '_connection_lost', '_drain_waiter', '_exit_waiters', '_extra', '_paused',
        '_pending_calls', '_subprocess_stderr_protocol', '_subprocess_stdin_protocol', '_subprocess_stdout_protocol',
        'closed', 'loop', 'process_id', 'process', 'return_code', 'stderr', 'stdin', 'stdout', )
    
    async def __new__(cls, loop, process_arguments, shell, stdin, stdout, stderr, buffer_size, extra,
            process_open_kwargs):
        """
        Creates a new ``AsyncProcess`` instance.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the async process is bound to.
        process_arguments : `tuple` of `Any`
            Process arguments to open the subprocess with.
        shell : `bool
            Whether the specified command will be executed through the shell.
        stdin : `file-like`, `subprocess.PIPE`.
            Standard input for the created shell. Defaults to `subprocess.PIPE`.
        stdout : `file-like`, `subprocess.PIPE`, `subprocess.DEVNULL`
            Standard output for the created shell.
        stderr : `file-like`, `subprocess.PIPE`, `subprocess.DEVNULL`, `subprocess.STDOUT`
            Standard error for the created shell.
        buffer_size : `int`
            Will be supplied as the corresponding argument to the open() function when creating the
            stdin/stdout/stderr pipe file objects:
            
            Expected values:
            
            +---------------+-----------+-----------------------------------------------------------------------+
            | Name          | Value     | Description                                                           |
            +===============+===========+=======================================================================+
            | unbuffered    | `0`       | Read and write are one system call and can return short.              |
            +---------------+-----------+-----------------------------------------------------------------------+
            | line buffered | `1`       | Only usable if `universal_newlines=True`, for example in text mode.   |
            +---------------+-----------+-----------------------------------------------------------------------+
            | buffer size   | `>1`      | Use a buffer of approximately to that value.                          |
            +---------------+-----------+-----------------------------------------------------------------------+
            | default       | `<0`      | use the system default: `io.DEFAULT_BUFFER_SIZE`.                     |
            +---------------+-----------+-----------------------------------------------------------------------+
            
        extra : `None` or `dict` of (`str`, `Any`) items
            Optional transport information.
        process_open_kwargs : `dict` of (`str`, `Any`) items
            Additional parameters to open the process with.
        
        Raises
        ------
        TypeError
            If `process_open_kwargs` contains unexpected key.
        """
        if stdin == PIPE:
            # Use a socket pair for stdin, since not all platforms support selecting read events on the write end of a
            # socket (which we use in order to detect closing of the other end).  Notably this is needed on AIX, and
            # works just fine on other platforms.
            stdin_r, stdin_w = create_socket_pair()
        else:
            stdin_r = stdin
            stdin_w = None
        
        process = None
        
        try:
            process = Popen(process_arguments, shell=shell, stdin=stdin_r, stdout=stdout, stderr=stderr,
                universal_newlines=False, bufsize=buffer_size, **process_open_kwargs)
            
            if (stdin_w is not None):
                stdin_r.close()
                process.stdin = open(stdin_w.detach(), 'wb', buffering=buffer_size)
                stdin_w = None
        except:
            if (process is not None) and (process.poll() is None):
                try:
                    process.kill()
                except ProcessLookupError:
                    pass
            
            raise
        finally:
            if (stdin_w is not None):
                stdin_r.close()
                stdin_w.close()
        
        if extra is None:
            extra = {}
        
        extra['subprocess'] = process
        
        self = object.__new__(cls)
        self._extra = extra
        self.closed = False
        self.loop = loop
        self.process = process
        self.process_id = process.pid
        self.return_code = None
        self._exit_waiters = None
        self._pending_calls = []
        self._subprocess_stdin_protocol = None
        self._subprocess_stdout_protocol = None
        self._subprocess_stderr_protocol = None
        self.stdin = None
        self.stdout = None
        self.stderr = None
        self._paused = False
        self._drain_waiter = None
        self._connection_lost = False
        self._alive_fds = []
        
        try:
            stdin = process.stdin
            if (stdin is not None):
                subprocess_stdin_protocol = SubprocessWritePipeProtocol(self, 0)
                await loop.connect_write_pipe(subprocess_stdin_protocol, stdin)
                self._subprocess_stdin_protocol = subprocess_stdin_protocol
                stdin_transport = subprocess_stdin_protocol.transport
                if (stdin_transport is not None):
                    self.stdin = SubprocessStreamWriter(loop, stdin_transport, self)
            
            stdout = process.stdout
            if (stdout is not None):
                subprocess_stdout_protocol = SubprocessReadPipeProtocol(self, 1)
                await loop.connect_read_pipe(subprocess_stdout_protocol, stdout)
                self._subprocess_stdout_protocol = subprocess_stdout_protocol
                stdout_transport = subprocess_stdout_protocol.transport
                if (stdout_transport is not None):
                    self.stdout = stdout_protocol = ReadProtocolBase(loop)
                    stdout_protocol.connection_made(stdout_transport)
                    self._alive_fds.append(1)
            
            stderr = process.stderr
            if (stderr is not None):
                subprocess_stderr_protocol = SubprocessReadPipeProtocol(self, 2)
                await loop.connect_read_pipe(subprocess_stderr_protocol, stderr)
                self._subprocess_stderr_protocol = subprocess_stderr_protocol
                stderr_transport = subprocess_stderr_protocol.transport
                if (stderr_transport is not None):
                    self.stderr = ReadProtocolBase(loop)
                    self.stderr.connection_made(stderr_transport)
                    self._alive_fds.append(2)
            
            for pending_call, args in self._pending_calls:
                pending_call(self, *args)
            
            self._pending_calls = None
        except:
            self.close()
            await self.wait()
            raise
        
        return self
    
    def __repr__(self):
        """Returns the async process's representation."""
        result = [
            '<',
            self.__class__.__name__,
                ]
        
        if self.closed:
            result.append(' closed')
            add_comma = True
        else:
            add_comma = False
        
        stdin = self.stdin
        if (stdin is not None):
            if add_comma:
                result.append(',')
            else:
                add_comma = True
            
            result.append(' stdin=')
            result.append(repr(stdin))
        
        stdout = self.stdout
        if (stdout is not None):
            if add_comma:
                result.append(',')
            else:
                add_comma = True
            
            result.append(' stdout=')
            result.append(repr(stdout))
        
        stderr = self.stderr
        if (stderr is not None):
            if add_comma:
                result.append(',')
            
            result.append(' stderr=')
            result.append(repr(stderr))
        
        result.append('>')
        
        return ''.join(result)
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self._extra.get(name, default)
    
    def is_closing(self):
        """
        Returns whether the async process is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self.closed
    
    def close(self):
        """
        Starts the shutdown process of the async process.
        """
        if self.closed:
            return
        
        self.closed = True
        
        pipe_stdin = self._subprocess_stdin_protocol
        if (pipe_stdin is not None):
            pipe_stdin.transport.close()
        
        pipe_stdout = self._subprocess_stdout_protocol
        if (pipe_stdout is not None):
            pipe_stdout.transport.close()
        
        pipe_stderr = self._subprocess_stderr_protocol
        if (pipe_stderr is not None):
            pipe_stderr.transport.close()
        
        process = self.process
        if (process is not None):
            if (self.return_code is None) and (process.poll() is None):
                try:
                    process.kill()
                except ProcessLookupError:
                    pass
        
        Task(self._process_exited(), self.loop)
    
    __del__ = close
    
    def send_signal(self, signal):
        """
        Sends the signal to the child process.
        
        Parameters
        ----------
        signal : `int`
            The signal to send.
        
        Raises
        ------
        ProcessLookupError
            The underlying process is already dead.
        """
        process = self.process
        if process is None:
            raise ProcessLookupError()
        
        process.send_signal(signal)
    
    def terminate(self):
        """
        Stops the child process.
        
        Raises
        ------
        ProcessLookupError
            The underlying process is already dead.
        """
        process = self.process
        if process is None:
            raise ProcessLookupError()
        
        process.terminate()
    
    async def kill(self):
        """
        Kills the child process.
        
        This method is a coroutine.
        """
        process = self.process
        if process is None:
            return
        
        process.kill()
        
        self.close()
        
        loop = self.loop
        future = Future(loop)
        loop.call_later(0.0, Future.set_result_if_pending, future, None)
        await future
    
    def _pipe_connection_lost(self, fd, exception):
        """
        Called when a file descriptor of the subprocess lost connection.
        
        Calls ``._do_pipe_connection_lost`` or adds it as a callback if the process is still connecting.
        
        Parameters
        ----------
        fd : `int`
            File descriptor identifier.
            
            It's value can be any of the following:
            
            +-------------------+-------+
            | Respective name   | Value |
            +===================+=======+
            | stdin             | `0`   |
            +-------------------+-------+
            | stdout            | `1`   |
            +-------------------+-------+
            | stderr            | `2`   |
            +-------------------+-------+
        
        exception : `None` or `BaseException` instance
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._do_pipe_connection_lost(fd, exception)
        else:
            pending_calls.append((self.__class__._do_pipe_connection_lost, (fd, exception)))
        
        self._try_finish()
    
    def _do_pipe_connection_lost(self, fd, exception):
        """
        Called by ``._pipe_connection_lost`` to call the pipe's connection lost method.
        
        Parameters
        ----------
        fd : `int`
            File descriptor identifier.
            
            It's value can be any of the following:
            
            +-------------------+-------+
            | Respective name   | Value |
            +===================+=======+
            | stdin             | `0`   |
            +-------------------+-------+
            | stdout            | `1`   |
            +-------------------+-------+
            | stderr            | `2`   |
            +-------------------+-------+
        
        exception : `None` or `BaseException` instance
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        if fd == 0:
            pipe = self.stdin
            if (pipe is not None):
                pipe.close()
            
            self._do_connection_lost(exception)
            return
        
        if fd == 1:
            reader = self.stdout
        elif fd == 2:
            reader = self.stderr
        else:
            reader = None
        
        if (reader is not None):
            reader.connection_lost(exception)
        
        try:
            self._alive_fds.remove(fd)
        except ValueError:
            pass
        
        self._maybe_process_exited()
    
    def _pipe_data_received(self, fd, data):
        """
        Called when one of the subprocess's pipe receives any data.
        
        Calls ``._do_pipe_data_received`` or adds it as a callback if the process is still connecting.
        
        Parameters
        ----------
        fd : `int`
            File descriptor identifier.
            
            It's value can be any of the following:
            
            +-------------------+-------+
            | Respective name   | Value |
            +===================+=======+
            | stdin             | `0`   |
            +-------------------+-------+
            | stdout            | `1`   |
            +-------------------+-------+
            | stderr            | `2`   |
            +-------------------+-------+
        
        data : `bytes`
            The received data.
        """
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._do_pipe_data_received(fd, data)
        else:
            pending_calls.append((self.__class__._do_pipe_data_received, (fd, data)))
    
    def _do_pipe_data_received(self, fd, data):
        """
        Called by ``._pipe_data_received`` to call the respective protocol' data received method.
        
        Parameters
        ----------
        fd : `int`
            File descriptor identifier.
            
            It's value can be any of the following:
            
            +-------------------+-------+
            | Respective name   | Value |
            +===================+=======+
            | stdin             | `0`   |
            +-------------------+-------+
            | stdout            | `1`   |
            +-------------------+-------+
            | stderr            | `2`   |
            +-------------------+-------+
        
        data : `bytes`
            The received data.
        """
        if fd == 1:
            reader = self.stdout
        elif fd == 2:
            reader = self.stderr
        else:
            return
        
        if (reader is not None):
            reader.data_received(data)
    
    async def _process_exited(self):
        """
        Task created by ``.close``to wait till the sub-process is closed and to set
        
        This method is a coroutine.
        """
        try:
            return_code = self.process.poll()
            if return_code is None:
                return_code = await self.loop.run_in_executor(self.process.wait)
            
            self.return_code = return_code
            
            self._try_finish()
            
            # wake up futures waiting for wait()
            exit_waiters = self._exit_waiters
            if (exit_waiters is not None):
                for waiter in self._exit_waiters:
                    if not waiter.cancelled():
                        waiter.set_result(return_code)
            
            self._exit_waiters = None
        finally:
            self.process = None
    
    def _maybe_process_exited(self):
        """
        Called when a sub-process pipe is closed. When all all pipe is closed, calls ``.close``.
        """
        if self._alive_fds:
            return
        
        self.close()
    
    async def wait(self, timeout=None):
        """
        Wait for child process to terminate.
        
        This method is a coroutine.
        
        Parameters
        ----------
        timeout : `None` or `float`
            The maximal amount of time to wait for the process to close in seconds.

        Returns
        -------
        return_code : `int`
            The returncode of the subprocess.
        
        Raises
        ------
        TimeoutExpired
            If the process was not closed before timeout.
        """
        return_code = self.return_code
        if (return_code is not None):
            return return_code
        
        waiter = Future(self.loop)
        
        exit_waiters = self._exit_waiters
        if exit_waiters is None:
            self._exit_waiters = exit_waiters = set()
        
        exit_waiters.add(waiter)
        if (timeout is not None):
            future_or_timeout(waiter, timeout)
        
        try:
            return await waiter
        except TimeoutError:
            exit_waiters.discard(waiter)
            
            process = self.process
            if process is None:
                args = None
            else:
                args = process.args
            
            raise TimeoutExpired(args, timeout) from None
    
    def _try_finish(self):
        """
        If the sub-process finished closing, calls ``._do_connection_lost``. If the process is still connecting, adds
        it as a callback instead.
        """
        if self.return_code is None:
            return
        
        stdin_protocol = self._subprocess_stdin_protocol
        if (stdin_protocol is None) or stdin_protocol.disconnected:
            return
        
        stdout_protocol = self._subprocess_stdout_protocol
        if (stdout_protocol is None) or stdout_protocol.disconnected:
            return
        
        stderr_protocol = self._subprocess_stderr_protocol
        if (stderr_protocol is None) or stderr_protocol.disconnected:
            return
        
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._do_connection_lost(None)
        else:
            pending_calls.append((self.__class__._do_connection_lost, (self, None,)))
    
    async def _feed_stdin(self, input_):
        """
        Feeds the given data to ``.stdin``, waits till it drains and closes it.
        
        Used by ``.communicate``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        input_ : `None` or `bytes-like`
            Optional data to be sent to the sub-process.
        """
        stdin = self.stdin
        if stdin is None:
            return
        
        if (input_ is not None):
            stdin.write(input_)
        
        try:
            await stdin.drain()
        except (BrokenPipeError, ConnectionResetError):
            # communicate() ignores BrokenPipeError and ConnectionResetError
            pass
        
        stdin.close()
    
    async def _read_close_stdout_stream(self):
        """
        Reads every data from ``.stdout``. When reading is done, closes it.
        
        Used by ``.communicate``.
        
        This method is a coroutine.
        
        Returns
        -------
        result :`None` or `bytes`
        """
        stream = self.stdout
        if stream is None:
            return None
        
        transport = self._subprocess_stdout_protocol.transport
        result = await stream.read()
        transport.close()
        return result
    
    async def _read_close_stderr_stream(self):
        """
        Reads every data from ``.stderr``. When reading is done, closes it.
        
        Used by ``.communicate``.
        
        This method is a coroutine.
        
        Returns
        -------
        result : `None` or `bytes`
        """
        stream = self.stderr
        if stream is None:
            return None
        
        transport = self._subprocess_stderr_protocol.transport
        result = await stream.read()
        transport.close()
        return result
    
    async def communicate(self, input_=None, timeout=None):
        """
        Sends data to stdin and reads data from stdout and stderr.
        
        Returns when the process is closed or raises when timeout occurs.
        
        This method is a coroutine.
        
        Parameters
        ----------
        input_ : `None` or `bytes-like`, Optional
            Optional data to be sent to the sub-process.
        timeout : `None` or `float`, Optional
            The maximal amount of time to wait for the process to close in seconds.
        
        Returns
        -------
        stdout : `None` or `bytes`
            The read data from stdout.
        stderr : `None` or `bytes`
            The read data from stderr.
        
        Raises
        ------
        TimeoutExpired
            If the process was not closed before timeout.
        """
        tasks = []
        
        loop = self.loop
        if input_ is None:
            stdin_task = None
        else:
            stdin_task = Task(self._feed_stdin(input_), loop)
            tasks.append(stdin_task)
        
        if self.stdout is None:
            stdout_task = None
        else:
            stdout_task = Task(self._read_close_stdout_stream(), loop)
            tasks.append(stdout_task)
        
        if self.stderr is None:
            stderr_task = None
        else:
            stderr_task = Task(self._read_close_stderr_stream(), loop)
            tasks.append(stderr_task)
        
        if tasks:
            future = WaitTillAll(tasks, loop)
            if (timeout is not None):
                future_or_timeout(future, timeout)
            
            done, pending = await future
            
            if pending:
                # timeout occurred, cancel the read tasks and raise TimeoutExpired.
                for task in pending:
                    task.cancel()
                
                process = self.process
                if process is None:
                    args = None
                else:
                    args = process.args
                
                raise TimeoutExpired(args, timeout)
            
            if (stdin_task is not None):
                stdin_task.result()
            
            if (stdout_task is None):
                stdout = None
            else:
                stdout = stdout_task.result()
            
            if (stderr_task is None):
                stderr = None
            else:
                stderr = stderr_task.result()
        else:
            stdout = None
            stderr = None
        
        await self.wait()
        
        return stdout, stderr
    
    def poll(self):
        """
        Returns the subprocess's returncode if terminated.
        
        Returns
        -------
        return_code : `None` or `int`
        """
        return self.return_code
    
    def pause_writing(self):
        """
        Called when the write transport's buffer goes over the high-water mark.
        
        ``.pause_writing`` is called when the write buffer goes over the high-water mark, and eventually
        ``.resume_writing`` is called when the write buffer size reaches the low-water mark.
        """
        self._paused = True
    
    def resume_writing(self):
        """
        Called when the transport's write buffer drains below the low-water mark.
        
        See ``.pause_writing`` for details.
        """
        self._paused = False
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        drain_waiter.set_result_if_pending(None)
    
    def _do_connection_lost(self, exception):
        """
        Sets ``._connection_lost`` as `True` and set ``._drain_waiter`` result or exception as well.
        
        Parameters
        ----------
        exception : `None` or `BaseException` instance
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        self._connection_lost = True
        # Wake up the writer if currently paused.
        if not self._paused:
            return
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        if exception is None:
            drain_waiter.set_result_if_pending(None)
        else:
            drain_waiter.set_exception_if_pending(exception)
    
    async def _drain_helper(self):
        """
        Called when the transport buffer after writing goes over the high limit mark to wait till it is drained.
        
        This method is a coroutine.
        """
        if self._connection_lost:
            raise ConnectionResetError('Connection lost')
        
        if not self._paused:
            return
        
        self._drain_waiter = drain_waiter = Future(self)
        await drain_waiter



# -*- coding: utf-8 -*-
__all__ = ('CallableAnalyzer', )

from .utils import function, MethodLike

is_coroutine_function = NotImplemented
is_coroutine_generator_function = NotImplemented

CO_OPTIMIZED   = 1
CO_NEWLOCALS   = 2
CO_VARARGS     = 4
CO_VARKEYWORDS = 8
CO_NESTED      = 16
CO_GENERATOR   = 32
CO_NOFREE      = 64

CO_COROUTINE           = 128
CO_ITERABLE_COROUTINE  = 256
CO_ASYNC_GENERATOR     = 512
# matches `async def` functions and `@coroutine` functions.
CO_COROUTINE_ALL       = CO_COROUTINE|CO_ITERABLE_COROUTINE

INSTANCE_TO_ASYNC_FALSE           = 0
INSTANCE_TO_ASYNC_TRUE            = 1
INSTANCE_TO_ASYNC_CANNOT          = 2
INSTANCE_TO_ASYNC_GENERATOR_FALSE = 3
INSTANCE_TO_ASYNC_GENERATOR_TRUE  = 4

ARGUMENT_POSITIONAL_ONLY        = 0
ARGUMENT_POSITIONAL_AND_KEYWORD = 1
ARGUMENT_KEYWORD_ONLY           = 2
ARGUMENT_ARGS                   = 3
ARGUMENT_KWARGS                 = 4

class Argument:
    """
    Represents a callable's argument.
    
    Attributes
    ----------
    annotation : `Any`
        The argument's annotation if applicable. Defaults to `None`.
    default : `Any`
        The default value of the argument if applicable. Defaults to `None`.
    has_annotation : `bool`
        Whether the argument has annotation.
    has_default : `bool`
        Whether the argument has default value.
    name : `str`
        The argument's name.
    positionality : `int`
        Whether the argument is positional, keyword or such.
        
        Can be set one of the following:
        +-----------------------------------+-----------+
        | Respective Name                   | Value     |
        +===================================+===========+
        | ARGUMENT_POSITIONAL_ONLY          | 0         |
        +-----------------------------------+-----------+
        | ARGUMENT_POSITIONAL_AND_KEYWORD   | 1         |
        +-----------------------------------+-----------+
        | ARGUMENT_KEYWORD_ONLY             | 2         |
        +-----------------------------------+-----------+
        | ARGUMENT_ARGS                     | 3         |
        +-----------------------------------+-----------+
        | ARGUMENT_KWARGS                   | 4         |
        +-----------------------------------+-----------+
    reserved : `bool`
        Whether the argument is reserved.
        
        For example at the case of methods, the first argument is reserved for the `self` argument.
    """
    __slots__ = ('annotation', 'default', 'has_annotation', 'has_default', 'name', 'positionality', 'reserved', )
    
    def __repr__(self):
        """Returns the argument's representation."""
        result = []
        result.append('<')
        result.append(self.__class__.__name__)
        
        if self.reserved:
            result.append(' reserved ,')
        
        result.append(('positional only', ' positional', ' keyword only', ' args', ' kwargs')[self.positionality])
        
        result.append(', name=')
        result.append(repr(self.name))
        
        if self.has_default:
            result.append(', default=')
            result.append(repr(self.default))
        
        if self.has_annotation:
            result.append(', annotation=')
            result.append(repr(self.annotation))
        
        result.append('>')
        return ''.join(result)
    
    def is_positional_only(self):
        """
        Returns whether the argument is positional only.
        
        Returns
        -------
        is_positional_only : `bool`
        """
        positionality = self.positionality
        if positionality == ARGUMENT_POSITIONAL_ONLY:
            return True
        
        return False
    
    def is_positional(self):
        """
        Returns whether the argument is positional.
        
        Returns
        -------
        is_positional : `bool`
        """
        positionality = self.positionality
        if positionality == ARGUMENT_POSITIONAL_ONLY:
            return True
        
        if positionality == ARGUMENT_POSITIONAL_AND_KEYWORD:
            return True
        
        return False
    
    def is_keyword(self):
        """
        Returns whether the argument can be used as a keyword argument.
        
        Returns
        -------
        is_keyword : `bool`
        """
        positionality = self.positionality
        if positionality == ARGUMENT_POSITIONAL_AND_KEYWORD:
            return True
        
        if positionality == ARGUMENT_KEYWORD_ONLY:
            return True
        
        return False
    
    def is_keyword_only(self):
        """
        Returns whether they argument is keyword only.
        
        Returns
        -------
        is_keyword_only : `bool`
        """
        positionality = self.positionality
        if positionality == ARGUMENT_KEYWORD_ONLY:
            return True
        
        return False
    
    def is_args(self):
        """
        Returns whether the argument is an `*args` argument.
        
        Returns
        -------
        is_args : `bool`
        """
        positionality = self.positionality
        if positionality == ARGUMENT_ARGS:
            return True
        
        return False
    
    def is_kwargs(self):
        """
        Returns whether the argument is an `**kwargs` argument.
        
        Returns
        -------
        is_kwargs : `bool`
        """
        positionality = self.positionality
        if positionality == ARGUMENT_KWARGS:
            return True
        
        return False

class CallableAnalyzer:
    """
    Analyzer for callable-s.
    
    Can analyze functions, methods, callable objects and types or such.
    
    Attributes
    ----------
    args_argument : `None` or ``Argument``
        If the analyzed callable has `*args` argument, then this attribute is set to it. Defaults to `None`.
    arguments : `list` of ``Argument``
        The analyzed callable's arguments.
    callable : `callable`
        The analyzed object.
    instance_to_async : `int`
        Whether the analyzed object can be instanced to async.
        
        +---------------------------+-----------+-------------------------------------------+
        | Respective Name           | Value     | Description                               |
        +===========================+===========+===========================================+
        | INSTANCE_TO_ASYNC_FALSE   | 0         | Whether the object is async.              |
        +---------------------------+-----------+-------------------------------------------+
        | INSTANCE_TO_ASYNC_TRUE    | 1         | Whether the object is on async callable,  |
        |                           |           | but after instancing it, returns one.     |
        +---------------------------+-----------+-------------------------------------------+
        | INSTANCE_TO_ASYNC_CANNOT  | 2         | Whether the object is not async.          |
        +---------------------------+-----------+-------------------------------------------+
    kwargs_argument : `None` or ``Argument``
        If the analyzed callable has `**kwargs`, then this attribute is set to it. Defaults to `None`.
    method_allocation : `int`
        How much argument is allocated if the analyzed callable is method if applicable.
    real_function : `callable`
        The function wrapped by the given callable.
    """
    __slots__ = ('args_argument', 'arguments', 'callable', 'instance_to_async', 'kwargs_argument', 'method_allocation',
        'real_function', )
    
    def __repr__(self):
        """Returns the callable analyzer's representation."""
        result = []
        result.append('<')
        result.append(self.__class__.__name__)
        
        if self.is_async():
            result.append(' async')
        elif self.is_async_generator():
            result.append(' async generator')
        elif self.can_instance_to_async_callable():
            result.append(' instance async')
        elif self.can_instance_to_async_generator():
            result.append(' instance async generator')
        
        method_allocation = self.method_allocation
        if method_allocation:
            result.append(' method')
            if method_allocation!=1:
                result.append(' (')
                result.append(repr(method_allocation))
                result.append(')')
        
        result.append(' ')
        callable_ = self.callable
        result.append(repr(callable_))
        real_function = self.real_function
        if (callable_ is not real_function):
            result.append(' (')
            result.append(repr(real_function))
            result.append(')')
        
        result.append(', arguments=')
        result.append(repr(self.arguments))
        
        args_argument = self.args_argument
        if (args_argument is not None):
            result.append(', args=')
            result.append(repr(args_argument))
        
        kwargs_argument = self.kwargs_argument
        if (kwargs_argument is not None):
            result.append(', kwargs=')
            result.append(repr(kwargs_argument))
        
        result.append('>')
        return ''.join(result)
    
    def __new__(cls, callable_, as_method=False):
        """
        Analyzes the given callable.
        
        Parameters
        ----------
        callable_ : `callable`
            The callable to analyze.
        as_method : `bool`, Optional
            Whether the given `callable` is given as a `function`, but it should be analyzed as a `method`. Defaults
            to `False`.
        
        Raises
        ------
        TypeError
            If the given object is not callable, or could not be used as probably intended.
        """
        while True:
            if isinstance(callable_, function):
                
                real_function = callable_
                if is_coroutine_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_FALSE
                elif is_coroutine_generator_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_FALSE
                else:
                    instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                
                method_allocation = 0
                break
            
            if isinstance(callable_, MethodLike):
                real_function = callable_
                
                if is_coroutine_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_FALSE
                elif is_coroutine_generator_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_FALSE
                else:
                    instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                
                method_allocation = MethodLike.get_reserved_argcount(callable_)
                break
            
            if not isinstance(callable_, type) and hasattr(type(callable_), '__call__'):
                real_function = type(callable_).__call__
                
                if is_coroutine_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_FALSE
                elif is_coroutine_generator_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_FALSE
                else:
                    instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                    
                if type(real_function) is function:
                    method_allocation = 1
                else:
                    method_allocation = MethodLike.get_reserved_argcount(real_function)
                
                break
            
            if not issubclass(callable_, type) and isinstance(callable_, type):
                
                while True:
                    real_function = callable_.__new__
                    if not callable(real_function):
                        raise TypeError(f'`{callable_!r}.__new__` should be callable, got `{real_function!r}`')
                    
                    if real_function is not object.__new__:
                        if is_coroutine_function(real_function):
                            instance_to_async = INSTANCE_TO_ASYNC_FALSE
                        elif is_coroutine_generator_function(real_function):
                            instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_FALSE
                        else:
                            if hasattr(callable_, '__call__'):
                                call = callable_.__call__
                                if is_coroutine_function(call):
                                    instance_to_async = INSTANCE_TO_ASYNC_TRUE
                                elif is_coroutine_generator_function(call):
                                    instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_TRUE
                                else:
                                    instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                            else:
                                instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                        
                        if type(real_function) is function:
                            method_allocation = 1
                        else:
                            method_allocation = MethodLike.get_reserved_argcount(real_function)
                        
                        break
                    
                    real_function = callable_.__init__
                    if not callable(real_function):
                        raise TypeError(f'`{callable_!r}.__init__` should be callable, got `{real_function!r}`')
                    
                    if real_function is not object.__init__:
                        if hasattr(callable_, '__call__'):
                            call = callable_.__call__
                            if is_coroutine_function(call):
                                instance_to_async = INSTANCE_TO_ASYNC_TRUE
                            elif is_coroutine_generator_function(call):
                                instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_TRUE
                            else:
                                instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                        else:
                            instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                        
                        if type(real_function) is function:
                            method_allocation = 1
                        else:
                            method_allocation = MethodLike.get_reserved_argcount(real_function)
                        
                        break
                    
                    real_function = None
                    method_allocation = 0
                    
                    if hasattr(callable_,'__call__'):
                        call = callable_.__call__
                        if is_coroutine_function(call):
                            instance_to_async = INSTANCE_TO_ASYNC_TRUE
                        elif is_coroutine_generator_function(call):
                            instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_TRUE
                        else:
                            instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                    else:
                        instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                    break
                
                break
            
            raise TypeError(f'Expected function, method or a callable object, got {callable_!r}.')
        
        if as_method and type(callable_) is function:
            method_allocation += 1
        
        if (real_function is not None) and ( not hasattr(real_function, '__code__')):
            raise TypeError(f'Expected function, got `{real_function!r}`')
        
        arguments = []
        if (real_function is not None):
            argument_count = real_function.__code__.co_argcount
            accepts_args = real_function.__code__.co_flags&CO_VARARGS
            keyword_only_argument_count = real_function.__code__.co_kwonlyargcount
            accepts_kwargs = real_function.__code__.co_flags&CO_VARKEYWORDS
            positional_only_argcount = getattr(real_function.__code__, 'co_posonlyargcount', 0)
            default_argument_values=real_function.__defaults__
            default_keyword_only_argument_values = real_function.__kwdefaults__
            annotations = getattr(real_function, '__annotations__', None)
            if (annotations is None):
                annotations = {}
            
            start=0
            end=argument_count
            argument_names = real_function.__code__.co_varnames[start:end]
            
            start=end
            end=start+keyword_only_argument_count
            keyword_only_argument_names = real_function.__code__.co_varnames[start:end]
            
            if accepts_args:
                args_name = real_function.__code__.co_varnames[end]
                end +=1
            else:
                args_name= None
            
            if accepts_kwargs:
                kwargs_name = real_function.__code__.co_varnames[end]
            else:
                kwargs_name = None
            
            names_to_defaults = {}
            if (default_argument_values is not None) and default_argument_values:
                argument_index = argument_count - len(default_argument_values)
                default_index = 0
                while argument_index < argument_count:
                    name = argument_names[argument_index]
                    default = default_argument_values[default_index]
                    
                    names_to_defaults[name] = default
                    
                    argument_index +=1
                    default_index +=1
            
            if (default_keyword_only_argument_values is not None) and default_keyword_only_argument_values:
                argument_index = keyword_only_argument_count - len(default_keyword_only_argument_values)
                while argument_index < keyword_only_argument_count:
                    name = keyword_only_argument_names[argument_index]
                    default = default_keyword_only_argument_values[name]
                    
                    names_to_defaults[name]=default
                    
                    argument_index+=1
            
            if (method_allocation>argument_count) and (args_name is None):
                raise TypeError(f'The passed object is a method like, but has not enough positional arguments: '
                    f'`{real_function!r}`.')
            
            index = 0
            while index < argument_count:
                argument = Argument()
                name = argument_names[index]
                argument.name = name
                
                try:
                    annotation = annotations[name]
                except KeyError:
                    argument.has_annotation = False
                    argument.annotation = None
                else:
                    argument.has_annotation = True
                    argument.annotation = annotation
                
                try:
                    default = names_to_defaults[name]
                except KeyError:
                    argument.has_default = False
                    argument.default = None
                else:
                    argument.has_default = True
                    argument.default = default
                
                if index<positional_only_argcount:
                    argument.positionality = ARGUMENT_POSITIONAL_ONLY
                else:
                    argument.positionality = ARGUMENT_POSITIONAL_AND_KEYWORD
                
                argument.reserved = (index<method_allocation)
                arguments.append(argument)
                index = index+1
            
            index = 0
            while index < keyword_only_argument_count:
                argument = Argument()
                name = keyword_only_argument_names[index]
                argument.name = name
                
                try:
                    annotation = annotations[name]
                except KeyError:
                    argument.has_annotation = False
                    argument.annotation = None
                else:
                    argument.has_annotation = True
                    argument.annotation = annotation
                
                try:
                    default = names_to_defaults[name]
                except KeyError:
                    argument.has_default = False
                    argument.default = None
                else:
                    argument.has_default = True
                    argument.default = default
                
                argument.positionality = ARGUMENT_KEYWORD_ONLY
                argument.reserved = False
                arguments.append(argument)
                index = index+1
            
            if args_name is None:
                args_argument = None
            else:
                args_argument = Argument()
                args_argument.name = args_name
                
                try:
                    annotation = annotations[args_name]
                except KeyError:
                    args_argument.has_annotation = False
                    args_argument.annotation = None
                else:
                    args_argument.has_annotation = True
                    args_argument.annotation = annotation

                args_argument.has_default = False
                args_argument.default = None
                args_argument.positionality = ARGUMENT_ARGS
                
                if method_allocation > argument_count:
                    args_argument.reserved = True
                else:
                    args_argument.reserved = False
                
            
            if kwargs_name is None:
                kwargs_argument = None
            else:
                kwargs_argument = Argument()
                kwargs_argument.name = kwargs_name
                try:
                    annotation = annotations[kwargs_name]
                except KeyError:
                    kwargs_argument.has_annotation = False
                    kwargs_argument.annotation = None
                else:
                    kwargs_argument.has_annotation = True
                    kwargs_argument.annotation = annotation
                
                kwargs_argument.has_default = False
                kwargs_argument.default = None
                kwargs_argument.positionality = ARGUMENT_KWARGS
                kwargs_argument.reserved = False
            
        else:
            args_argument = None
            kwargs_argument = None
        
        self = object.__new__(cls)
        self.arguments = arguments
        self.args_argument = args_argument
        self.kwargs_argument = kwargs_argument
        self.callable = callable_
        self.method_allocation = method_allocation
        self.real_function = real_function
        self.instance_to_async = instance_to_async
        return self
    
    def is_async(self):
        """
        Returns whether the analyzed callable is async.
        
        Returns
        -------
        is_async : `bool`
        """
        if self.instance_to_async == INSTANCE_TO_ASYNC_FALSE:
            return True
        
        return False
    
    def is_async_generator(self):
        """
        Returns whether the analyzed callable is an async generator.
        
        Returns
        is_async_generator : `bool`
        """
        if self.instance_to_async == INSTANCE_TO_ASYNC_GENERATOR_FALSE:
            return True
        
        return False
    
    def can_instance_to_async_callable(self):
        """
        Returns whether the analyzed callable can be instanced to async.
        
        Returns
        -------
        can_instance_to_async_callable : `bool`
        """
        if self.instance_to_async != INSTANCE_TO_ASYNC_TRUE:
            return False
        
        for argument in self.arguments:
            if argument.reserved:
                continue
            
            if argument.has_default:
                continue
            
            return False
        
        return True
    
    def can_instance_to_async_generator(self):
        """
        Returns whether the analyzed callable can be instanced to async.
        
        Returns
        -------
        can_instance_to_async_callable : `bool`
        """
        if self.instance_to_async != INSTANCE_TO_ASYNC_GENERATOR_TRUE:
            return False
        
        for argument in self.arguments:
            if argument.reserved:
                continue
            
            if argument.has_default:
                continue
            
            return False
        
        return True
    
    # call `.can_instance_async_callable` or `.can_instance_to_async_generator` before
    def instance(self):
        """
        Instances the analyzed callable.
        
        Should be called only after a ``.can_instance_async_callable`` call, if it returned `True`.
        
        Returns
        -------
        instance_to_async_callable : `Any`
        """
        return self.callable()
    
    def get_non_default_keyword_only_argument_count(self):
        """
        Returns the amount of non default keyword only arguments of the analyzed callable.
        
        Returns
        -------
        non_default_keyword_only_argument_count : `int`
        """
        count = 0
        for value in self.arguments:
            if not value.is_keyword_only():
                continue
            
            if value.has_default:
                break
            
            count += 1
            continue
        
        return count
    
    def get_non_reserved_positional_arguments(self):
        """
        Returns the non reserved positional arguments of the analyzed callable.
        
        Returns
        -------
        non_reserved_positional_arguments : `list` of ``Argument``
        """
        result = []
        for argument in self.arguments:
            if not argument.is_positional():
                break
            
            if argument.reserved:
                continue
            
            result.append(argument)
            continue
        
        return result
    
    def get_non_reserved_positional_argument_count(self):
        """
        Returns the amount of the non reserved positional arguments of the analyzed callable.
        
        Returns
        -------
        non_reserved_positional_arguments : `int`
        """
        count = 0
        for argument in self.arguments:
            if not argument.is_positional():
                break
            
            if argument.reserved:
                continue
            
            count +=1
            continue
        
        return count
    
    def get_non_reserved_non_default_argument_count(self):
        """
        Returns the amount of the non reserved non default arguments of the analyzed callable.
        
        Returns
        -------
        non_reserved_non_default_argument_count : `int`
        """
        count = 0
        for argument in self.arguments:
            if not argument.is_positional():
                break
            
            if argument.reserved:
                continue
            
            if argument.has_default:
                continue
            
            count +=1
            continue
        
        return count

    def get_non_reserved_positional_argument_range(self):
        """
        Returns the minimal and the maximal amount how much non reserved positional arguments the analyzed callable
        expects / accepts.
        
        Returns
        -------
        start : `int`
            The minimal amount of non reserved arguments, what the analyzed callable expects.
        end : `int`
            The maximal amount of non reserved arguments, what the analyzed callable accepts.
        
        Notes
        -----
        `*args` argument is ignored from the calculation.
        """
        iterator = iter(self.arguments)
        start = 0
        for argument in iterator:
            if not argument.is_positional():
                return start, start
            
            if argument.reserved:
                continue
            
            if argument.has_default:
                break
            
            start += 1
            continue
        
        else:
            return start, start
        
        end = start
        for argument in iterator:
            if not argument.is_positional():
                return start, end
            
            if argument.reserved:
                continue
            
            end += 1
            continue
        
        return start, end
    
    def accepts_args(self):
        """
        Returns whether the analyzed callable accepts `*args` argument.
        
        Returns
        -------
        accepts_args : `bool`
        """
        return (self.args_argument is not None)
    
    def accepts_kwargs(self):
        """
        Returns whether the analyzed callable accepts `**kwargs` argument.
        
        Returns
        -------
        accepts_kwargs : `bool`
        """
        return (self.kwargs_argument is not None)






# -*- coding: utf-8 -*-
__all__ = ('Closer', )

from ...backend.futures import Task
from ...discord.client_core import KOKORO
from ...discord.emoji import BUILTIN_EMOJIS
from ...discord.parsers import InteractionEvent
from ...discord.message import Message
from ...discord import ChannelTextBase
from ...discord.exceptions import DiscordException, ERROR_CODES

from .utils import GUI_STATE_READY, GUI_STATE_SWITCHING_PAGE, GUI_STATE_CANCELLING, GUI_STATE_CANCELLED, \
    GUI_STATE_SWITCHING_CTX, Timeouter


class Closer:
    """
    Familiar to ``Pagination``, but can be used if the given content has only 1 page, so only an `x` would show up.
    
    Picks up on reaction additions by any users.
    
    Attributes
    ----------
    canceller : `None` or `function`
        The function called when the ``Closer`` is cancelled or when it expires. This is a onetime use and after
        it was used, is set as `None`.
    channel : ``ChannelTextBase`` instance
        The channel where the ``Closer`` is executed.
    check : `None` or `callable`
        A callable what decides whether the ``Closer`` should process a received reaction event. Defaults to
        `None`.
        
        Should accept the following parameters:
        +-----------+---------------------------------------------------+
        | Name      | Type                                              |
        +===========+===================================================+
        | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
        +-----------+---------------------------------------------------+
        
        Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    client : ``Client`` of ``Embed`` (or any compatible)
        The client who executes the ``Closer``.
    message : `None` or ``Message``
        The message on what the ``Closer`` is executed.
    task_flag : `int`
        A flag to store the state of the ``Closer``.
        
        Possible values:
        +---------------------------+-------+-----------------------------------------------------------------------+
        | Respective name           | Value | Description                                                           |
        +===========================+=======+=======================================================================+
        | GUI_STATE_READY           | 0     | The closer does nothing, is ready to be used.                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_PAGE  | 1     | The closer is currently changing it's page.                           |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLING      | 2     | The pagination is currently changing it's page, but it was cancelled  |
        |                           |       | meanwhile.                                                            |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLED       | 3     | The pagination is, or is being cancelled right now.                   |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_CTX   | 4     | The closer is switching context. Not used by the default class,       |
        |                           |       | but expected.                                                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
    timeouter : `None` or ``Timeouter``
        Executes the timing out feature on the ``Closer``.
    
    Class Attributes
    ----------------
    CANCEL : ``Emoji`` = `BUILTIN_EMOJIS['x']`
        The emoji used to cancel the ``Closer``.
    """
    CANCEL = BUILTIN_EMOJIS['x']
    
    __slots__ = ('canceller', 'channel', 'check', 'client', 'message', 'task_flag', 'timeouter')
    
    async def __new__(cls, client, channel, content, timeout=240., message=None, check=None):
        """
        Creates a new pagination with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who will execute the ``Closer``.
        channel : ``ChannelTextBase`` instance, ``Message``, ``InteractionEvent``
            The channel where the ``Closer`` will be executed.  Pass it as a ``Message`` instance to send a reply.
        
            If given as ``InteractionEvent``, then will acknowledge it and create a new message with it as well.
            Although will not acknowledge it if `message` is given.
        
        content : ``Any`
            The displayed content.
        timeout : `float`, Optional
            The timeout of the ``Closer`` in seconds. Defaults to `240.0`.
        message : `None` or ``Message``, Optional
            The message on what the ``Closer`` will be executed. If not given a new message will be created.
            Defaults to `None`.
        check : `None` or `callable`, Optional
            A callable what decides whether the ``Closer`` should process a received reaction event. Defaults to
            `None`.
            
            Should accept the following parameters:
            +-----------+---------------------------------------------------+
            | Name      | Type                                              |
            +===========+===================================================+
            | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
            +-----------+---------------------------------------------------+
            
            Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
            
            Should return the following values:
            +-------------------+-----------+
            | Name              | Type      |
            +===================+===========+
            | should_process    | `bool`    |
            +-------------------+-----------+
        
        Returns
        -------
        self : `None` or ``Closer``
        
        Raises
        ------
        TypeError
            `channel`'s type is incorrect.
        """
        if isinstance(channel, ChannelTextBase):
            target_channel = channel
            received_interaction = False
        elif isinstance(channel, Message):
            target_channel = channel.channel
            received_interaction = False
        elif isinstance(channel, InteractionEvent):
            target_channel = channel.channel
            received_interaction = True
        else:
            raise TypeError(f'`channel` can be given only as `{ChannelTextBase.__name__}`, `{Message.__name__}` '
                f'of as {InteractionEvent.__name__} instance, got {channel.__class__.__name__}.')
        
        self = object.__new__(cls)
        self.check = check
        self.client = client
        self.channel = target_channel
        self.canceller = cls._canceller
        self.task_flag = GUI_STATE_READY
        self.message = message
        self.timeouter = None
        
        try:
            if message is None:
                if received_interaction:
                    if not channel.is_acknowledged():
                        await client.interaction_response_message_create(channel)
                    
                    message = await client.interaction_followup_message_create(channel, content)
                else:
                    message = await client.message_create(channel, content)
                self.message = message
            else:
                await client.message_edit(message, content)
            
            if not target_channel.cached_permissions_for(client).can_add_reactions:
                return self
            
            await client.reaction_add(message, self.CANCEL)
        except BaseException as err:
            if isinstance(err, ConnectionError):
                return None
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.max_reactions, # reached reaction 20, some1 is trolling us.
                        ERROR_CODES.invalid_access, # client removed
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                        ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                            ):
                    return None
            
            raise
        
        self.timeouter = Timeouter(self, timeout=timeout)
        client.events.reaction_add.append(message, self)
        return self
    
    
    async def __call__(self, client, event):
        """
        Called when a reaction is added on the respective message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who executes the ``Closer``
        event : ``ReactionAddEvent``
            The received event.
        """
        if event.user.is_bot:
            return
        
        if (event.emoji is not self.CANCEL):
            return
        
        task_flag = self.task_flag
        if task_flag != GUI_STATE_READY:
            # ignore GUI_STATE_SWITCHING_PAGE and GUI_STATE_CANCELLED and GUI_STATE_SWITCHING_CTX
            return
        
        check = self.check
        if (check is not None):
            try:
                should_continue = check(event)
            except BaseException as err:
                await client.events.error(client, f'{self!r}.__call__', err)
                return
            
            if not should_continue:
                return
        
        self.task_flag = GUI_STATE_CANCELLED
        try:
            await client.message_delete(self.message)
        except BaseException as err:
            self.cancel()
            
            if isinstance(err, ConnectionError):
                # no internet
                return
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.invalid_access, # client removed
                            ):
                    return
            
            await client.events.error(client, f'{self!r}.__call__', err)
            return
    
    async def _canceller(self, exception,):
        """
        Used when the ``Closer`` is cancelled.
        
        First removes the respective waitfors of the canceller, then set it's ``.task_flag`` of the it to
        `GUI_STATE_CANCELLED`.
        
        If `exception` is given as `TimeoutError`, then removes the ``Closer``'s reactions from the respective
        message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance
            Exception to cancel the ``Closer`` with.
        """
        client = self.client
        message = self.message
        
        client.events.reaction_add.remove(message, self)
        
        if self.task_flag == GUI_STATE_SWITCHING_CTX:
            # the message is not our, we should not do anything with it.
            return
        
        self.task_flag = GUI_STATE_CANCELLED
        
        if exception is None:
            return
        
        if isinstance(exception, TimeoutError):
            if self.channel.cached_permissions_for(client).can_manage_messages:
                try:
                    await client.reaction_clear(message)
                except BaseException as err:
                    
                    if isinstance(err, ConnectionError):
                        # no internet
                        return
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_message, # message deleted
                                ERROR_CODES.unknown_channel, # channel deleted
                                ERROR_CODES.invalid_access, # client removed
                                ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                                    ):
                            return
                    
                    await client.events.error(client, f'{self!r}._canceller', err)
                    return
            return
        
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
        #we do nothing
    
    def cancel(self, exception=None):
        """
        Cancels the closer, if it is not cancelled yet.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance, Optional
            Exception to cancel the closer with. Defaults to `None`
        """
        canceller = self.canceller
        if canceller is None:
            return
        
        self.canceller = None
        
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
        
        return Task(canceller(self, exception), KOKORO)
    
    def __repr__(self):
        """Returns the pagination's representation."""
        result = [
            '<', self.__class__.__name__,
            ' client=', repr(self.client),
            ', channel=', repr(self.channel),
            ', task_flag='
                ]
        
        task_flag = self.task_flag
        result.append(repr(task_flag))
        result.append(' (')
        
        task_flag_name = (
            'GUI_STATE_READY',
            'GUI_STATE_SWITCHING_PAGE',
            'GUI_STATE_CANCELLING',
            'GUI_STATE_CANCELLED',
            'GUI_STATE_SWITCHING_CTX',
                )[task_flag]
        
        result.append(task_flag_name)
        result.append(')>')
        
        return ''.join(result)



# -*- coding: utf-8 -*-
from ...backend.utils import multidict
from ...discord.parsers import asynclist, ChunkWaiter, EVENTS, DEFAULT_EVENT
from ...discord.client_core import CLIENTS
from ...discord.client_utils import WaitForHandler

SNAPSHOT_TAKERS = {}

# Client.events snapshot functions

IGNORED_EVENT_HANDLER_TYPES = {
    WaitForHandler,
    ChunkWaiter,
        }

def should_ignore_event_handler(event_handler):
    """
    Returns whether the given `event_handler` should be ignored from snapshotting.
    
    Parameters
    ----------
    event_handler : `async-callable`
        The respective event handler.
    
    Returns
    -------
    should_ignore : `bool`
    """
    if event_handler is DEFAULT_EVENT:
        return True
    
    if type(event_handler) in IGNORED_EVENT_HANDLER_TYPES:
        return True
    
    return False

EVENT_NAMES = tuple(EVENTS.defaults)


def take_event_handler_snapshot(client):
    """
    Collects the event handlers of the client.
    
    Parameters
    ----------
    client : ``Client``
        The client, who will be snapshotted.
    
    Returns
    -------
    collected : `multidict` of `str`, `async-callable` items
        A multidict storing `event-name`, `event-handler` pairs.
    """
    collected = multidict()
    
    event_descriptor = client.events
    for event_name in EVENT_NAMES:
        event_handler = getattr(event_descriptor, event_name)
        if isinstance(event_handler, asynclist):
            for event_handler in event_handler:
                if should_ignore_event_handler(event_handler):
                    continue
                
                collected[event_name] = event_handler
        else:
            if should_ignore_event_handler(event_handler):
                continue
            
            collected[event_name] = event_handler
    
    return collected


def calculate_event_handler_snapshot_difference(client, snapshot_old, snapshot_new):
    """
    Calculates the difference between two event handler snapshot returning the difference.
    
    Parameters
    ----------
    client : ``Client``
        The respective client instance.
    snapshot_old : `multidict` of `str`, `async-callable` items
        An old snapshot taken.
    snapshot_new : `multidict` of `str`, `async-callable` items
        A new snapshot.
    
    Returns
    -------
    snapshot_difference : `None` or `list` of `tuple` \
            (`str`, (`None` or `list` of `async-callable`), (`None` or `list` of `async-callable`))
        A list of event handler differences in a list of tuples.
        
        The tuple has 3 elements, where the 0th element is the name of the respective event, meanwhile the 1th element
        contains the removed event handlers and the 2th the added ones.
        
        If there is no difference between two snapshots, returns `None`.
    """
    snapshot_difference = []
    
    event_names = set(snapshot_old)
    event_names.update(snapshot_new)
    
    for event_name in event_names:
        old_handlers = snapshot_old.get_all(event_name)
        new_handlers = snapshot_new.get_all(event_name)
        
        if (old_handlers is not None) and (new_handlers is not None):
            for index in reversed(range(len(old_handlers))):
                handler = old_handlers[index]
                try:
                    new_handlers.remove(handler)
                except ValueError:
                    pass
                else:
                    del old_handlers[index]
            
            if not new_handlers:
                new_handlers = None
            
            if not old_handlers:
                old_handlers = None
        
        if (old_handlers is not None) or (new_handlers is not None):
            snapshot_difference.append((event_name, old_handlers, new_handlers))
    
    if not snapshot_difference:
        snapshot_difference = None
    
    return snapshot_difference


def revert_event_handler_snapshot(client, snapshot_difference):
    """
    Reverts a taken snapshot.
    
    Parameters
    ----------
    client : ``Client``
        The respective client instance.
    snapshot_difference : `list` of `tuple` \
            (`str`, (`None` or `list` of `async-callable`), (`None` or `list` of `async-callable`))
        A list of event handler differences in a list of tuples.
        
        The tuple has 3 elements, where the 0th element is the name of the respective event, meanwhile the 1th element
        contains the removed event handlers and the 2th the added ones.
    """
    event_descriptor = client.events
    for event_handler_name, removed_handlers, added_handlers in snapshot_difference:
        if (added_handlers is not None):
            for handler in added_handlers:
                event_descriptor.remove(handler, name=event_handler_name, count=1)
            
        if (removed_handlers is not None):
            for handler in removed_handlers:
                event_descriptor(handler, name=event_handler_name)


# Register snapshot taker
SNAPSHOT_TAKERS['client.events'] = (
    take_event_handler_snapshot,
    calculate_event_handler_snapshot_difference,
    revert_event_handler_snapshot,
        )


# General snapshot functions

def take_snapshot():
    """
    Takes a snapshot of the clients.
    
    Returns
    -------
    snapshot : `list` of `tuple` (``Client``, `list` of `tuple` (`str`, `Any`))
        A taken snapshot. Each element of the generated list contains a `client`, `client-snapshot` pair, where
        the `client-snapshot` is a list containing `snapshot-type-name`, `type-specific-snapshot` pairs.
    """
    snapshot = []
    for client in CLIENTS:
        client_snapshot = []
        snapshot.append((client, client_snapshot))
        
        for snapshot_type_name, (snapshot_taker, difference_calculator, reverter) in SNAPSHOT_TAKERS.items():
            sub_snapshot = snapshot_taker(client)
            
            client_snapshot.append((snapshot_type_name, sub_snapshot))
    
    return snapshot

def calculate_snapshot_difference(snapshot_old, snapshot_new):
    """
    Calculates snapshot differences between two snapshots.
    
    Parameters
    ----------
    snapshot_old : `list` of `tuple` (``Client``, `list` of `tuple` (`str`, `Any`))
        Old taken snapshot.
    snapshot_new : `list` of `tuple` (``Client``, `list` of `tuple` (`str`, `Any`))
        New taken snapshot.
    
    Returns
    -------
    snapshot_difference : `list` of `tuple` (``Client``, `list` of `tuple` (`str`, `Any`))
        The difference between two snapshots. Each element of the generated list contains a `client`,
        `client-snapshot-difference` pair, where the `client-snapshot-difference` is a dictionary containing
        `snapshot-type-name`, `type-specific-snapshot-difference` pairs.
    """
    # First check client difference. We will ignore clients which weren't present before or after.
    snapshot_clients_interception = set(e[0] for e in snapshot_old)&set(e[0] for e in snapshot_new)
    
    # Place snapshots next to each other depending on clients.
    parallel_snapshots = []
    for present_client in snapshot_clients_interception:
        for client, client_snapshot in snapshot_old:
            if present_client is client:
                client_snapshot_old = client_snapshot
                break
        
        for client, client_snapshot in snapshot_new:
            if present_client is client:
                client_snapshot_new = client_snapshot
                break
        
        parallel_snapshots.append((present_client, client_snapshot_old, client_snapshot_new))
    
    snapshot_difference = []
    
    for client, client_snapshot_old, client_snapshot_new in parallel_snapshots:
        client_snapshot_difference = []
        snapshot_difference.append((client, client_snapshot_difference))
        
        # Collect snapshot type names, which are present.
        snapshot_types_name_interception = set(e[0] for e in client_snapshot_old)&set(e[0] for e in client_snapshot_new)
        
        client_snapshot_old = dict(client_snapshot_old)
        client_snapshot_new = dict(client_snapshot_new)
        
        for snapshot_type_name in snapshot_types_name_interception:
            type_specific_snapshot_old = client_snapshot_old[snapshot_type_name]
            type_specific_snapshot_new = client_snapshot_new[snapshot_type_name]
            snapshot_taker, difference_calculator, reverter = SNAPSHOT_TAKERS[snapshot_type_name]
            
            snapshot_type_specific_difference = \
                difference_calculator(client, type_specific_snapshot_old, type_specific_snapshot_new)
            
            # `difference_calculator` might return `None`
            if (snapshot_type_specific_difference is not None):
                client_snapshot_difference.append((snapshot_type_name, snapshot_type_specific_difference))
    
    return snapshot_difference


def revert_snapshot(snapshot_difference):
    """
    Reverts the clients to a previous state based on snapshots.
    
    Parameters
    ----------
    snapshot_difference : `list` of `tuple` (``Client``, `list` of `tuple` (`str`, `Any`))
        Difference between 2 snapshot.
    """
    for client, client_snapshot_difference in snapshot_difference:
        for snapshot_type_name, snapshot_type_specific_difference in client_snapshot_difference:
            snapshot_taker, difference_calculator, reverter = SNAPSHOT_TAKERS[snapshot_type_name]
            reverter(client, snapshot_type_specific_difference)
